const gql = require('graphql-tag')
const _ = require('lodash')

const sortobject = require('deep-sort-object')
const reportTimeout = 10
const util = require('./util')

const schema = gql`
  extend type Query {
    "Introspect GraphQL server itself"
    introspect: Introspect
  }

  "Introspect contains basic information about the GraphQL server itself"
  type Introspect {
    "Self name"
    name: String!

    "Self version"
    version: String!

    "Self setup"
    setup: Object!

    "Self report"
    report: Object

    "Self queries"
    queries: Object

    "Self duration in ms"
    duration: Int
  }

`


const resolvers = {
  Query: {
    introspect: (_source, _args, {dataSources} ) => {
      if (dataSources.introspect) {
        return {}
      }
    },
  },
  Introspect: {
    name(_source, _args, context) {
      if (context.dataSources.introspect) return context.dataSources.introspect.getName()
    },
    version(_source, _args, context) {
      if (context.dataSources.introspect) return context.dataSources.introspect.getVersion()
    },
    setup(_source, _args, context) {
      if (context.dataSources.introspect) return context.dataSources.introspect.getSetup()
    },
    report(_source, _args, context) {
      if (context.dataSources.introspect) return context.dataSources.introspect.getReport(context)
    },
    queries(_source, _args, context) {
      if (context.dataSources.introspect) return context.dataSources.introspect.getQueries(context)
    },
    duration(_source, _args, context) {
      if (context.dataSources.introspect) return context.dataSources.introspect.getDuration(context)
    },
  }
}

function waitForAllOperationsDone(context) {
  return new Promise((resolve) => {
    let timeoutID
    let eventHandlerSet = false
    const checkPendingOps = () => {
      if (context.pendingOps > 0) {
        if (!eventHandlerSet) {
          context.events.on('allOperationsDone', () => {
            clearTimeout(timeoutID)
            resolve()
          })
          eventHandlerSet = true
        }
        timeoutID = setTimeout( checkPendingOps, reportTimeout) /* ensures resolving after no reporting within time limit */
      } else {
        resolve()
      }
    }
    timeoutID = setTimeout( checkPendingOps, reportTimeout) /* ensures resolving after no reporting within time limit */
  })
}

class Introspect {
  constructor(obj) {
    this.set(obj)
  }

  set({name,version,setup}) {
    if (name) this.name = name
    if (version) this.version = version
    if (setup) this.setup = setup
  }

  getName() {
    return this.name
  }

  getVersion() {
    return this.version
  }

  getSetup() {
    return this.setup || {}
  }

  async getReport(context) {
    return waitForAllOperationsDone(context).then( () => sortobject(_.get(context,'res.introspect.report')) )
  }

  async getQueries(context) {
    return waitForAllOperationsDone(context).then( () => _.get(context,'res.introspect.queries') )
  }


  async getDuration(context) {
    return waitForAllOperationsDone(context).then( () => util.duration(context.start))
  }
}

module.exports = {
  schema,
  resolvers,
  datasource: Introspect
}
